@@ -12,6 +12,7 @@
 #include "gemmi/pirfasta.hpp"  // for read_pir_or_fasta
 #include "gemmi/mmread_gz.hpp" // for read_structure_gz
 #include "gemmi/select.hpp"    // for Selection
+#include "gemmi/pymol_select.hpp" // for select_atoms
 #include "gemmi/enumstr.hpp"   // for polymer_type_to_string
 #include "gemmi/calculate.hpp" // for parse_triplet_as_ftransform
 
@@ -62,7 +63,7 @@ struct ConvArg: public Arg {
 enum OptionIndex {
   FormatIn=AfterCifModOptions, FormatOut, CifStyle, AllAuth, BlockName,
   ExpandNcs, AsAssembly,
-  RemoveH, RemoveWaters, RemoveLigWat, TrimAla, Select, Remove, ApplySymop,
+  RemoveH, RemoveWaters, RemoveLigWat, TrimAla, Select, Remove, PymolSel, ApplySymop,
   Reframe, ShortTer, Linkid, Linkr, CopyRemarks, Minimal, ShortenCN, RenameChain,
   ShortenTLC, ChangeCcdCode, SetSeq, SiftsNum,
   Biso, BisoScale, AddTls, Anisou, AssignRecords,
@@ -151,6 +152,8 @@ const option::Descriptor Usage[] = {
     "  --select=SEL  \tOutput only the specified selection." },
   { Remove, 0, "", "remove", Arg::Required,
     "  --remove=SEL  \tRemove the specified selection." },
+  { PymolSel, 0, "", "pysel", Arg::Required,
+    "  --pysel=SEL  \t[PyMOL selection syntax] Output only the specified selection." },
   { ApplySymop, 0, "", "apply-symop", Arg::Required,
     "  --apply-symop=OP  \tApply operation, e.g. '-x,y+1/2,-z' or 'x,y,z+0.1'." },
   { Reframe, 0, "", "reframe", Arg::None,
@@ -307,6 +310,10 @@ void convert(gemmi::Structure& st,
   if (st.models.empty())
     gemmi::fail("No atoms in the input (", format_as_string(st.input_format), ") file. "
                 "Wrong file format?");
+  if (st.non_ascii_line != 0)
+    std::cerr << "WARNING: non-ASCII byte found in the PDB input on line "
+              << st.non_ascii_line
+              << ".\nWARNING: The fixed-column PDB format may be misinterpreted." << std::endl;
   if (st.ter_status == 'e')
     std::cerr << "WARNING: TER records in the input PDB are clearly where they shouldn't be."
                  "\nWARNING: Ignoring all TER records." << std::endl;
@@ -344,6 +351,11 @@ void convert(gemmi::Structure& st,
     gemmi::Selection(options[Select].arg).remove_not_selected(st);
   if (options[Remove])
     gemmi::Selection(options[Remove].arg).remove_selected(st);
+  if (options[PymolSel]) {
+    gemmi::FlatStructure fs(st);
+    gemmi::remove_not_selected(fs, options[PymolSel].arg);
+    st = fs.generate_structure();
+  }
   if (st.models.empty())
     gemmi::fail("all models got removed");
   if (options[ApplySymop]) {
@@ -469,7 +481,7 @@ void convert(gemmi::Structure& st,
   }
 
   HowToNameCopiedChain how = HowToNameCopiedChain::AddNumber;
-  if (output_type == CoorFormat::Pdb)
+  if (output_type == CoorFormat::Pdb || options[ShortenCN])
     how = HowToNameCopiedChain::Short;
   if (options[AsAssembly]) {
     gemmi::Logger logger{&gemmi::Logger::to_stderr, options[Verbose] ? 6 : 3};
@@ -585,21 +597,23 @@ int GEMMI_MAIN(int argc, char **argv) {
               << "..." << std::endl;
   try {
     gemmi::Structure st;
-    if (p.options[OldPdb]) {
+    if (in_type == CoorFormat::ChemComp) {
+      int which = 7;
+      // chemcomp:m or chemcomp:i
+      if (p.options[FormatIn].arg && std::strlen(p.options[FormatIn].arg) > 9)
+        // cf. ChemCompModel
+        which = p.options[FormatIn].arg[9] == 'i' ? 4 : 2;
+      st = gemmi::read_structure_from_chemcomp_gz(input, nullptr, which);
+    } else if (in_type == CoorFormat::Pdb || p.options[OldPdb] ||
+               (in_type == CoorFormat::Detect &&
+                gemmi::coor_format_from_ext_gz(input) == CoorFormat::Pdb)) {
       gemmi::PdbReadOptions options;
-      options.max_line_length = 72;
+      options.check_non_ascii = true;
+      if (p.options[OldPdb])
+        options.max_line_length = 72;
       st = gemmi::read_pdb_gz(input, options);
     } else {
-      if (in_type == CoorFormat::ChemComp) {
-        int which = 7;
-        // chemcomp:m or chemcomp:i
-        if (p.options[FormatIn].arg && std::strlen(p.options[FormatIn].arg) > 9)
-          // cf. ChemCompModel
-          which = p.options[FormatIn].arg[9] == 'i' ? 4 : 2;
-        st = gemmi::read_structure_from_chemcomp_gz(input, nullptr, which);
-      } else {
-        st = gemmi::read_structure_gz(input, in_type);
-      }
+      st = gemmi::read_structure_gz(input, in_type);
     }
     convert(st, output, out_type, p.options);
   } catch (std::runtime_error& e) {
