@@ -87,7 +87,7 @@ nb::class_<Grid<T>, GridBase<T>> add_grid_common(nb::module_& m, const std::stri
       new(grid) Gr();
       grid->set_size(nx, ny, nz);
     }, nb::arg("nx"), nb::arg("ny"), nb::arg("nz"))
-    .def("__init__", [](Gr* grid, const nb::ndarray<nb::numpy, T, nb::ndim<3>>& arr,
+    .def("__init__", [](Gr* grid, const nb::ndarray<nb::numpy, const T, nb::ndim<3>>& arr,
                         const UnitCell *cell, const SpaceGroup* sg) {
       new(grid) Gr();
       auto r = arr.view();
@@ -180,7 +180,7 @@ void add_grid_interpolation(nb::class_<Grid<T>, GridBase<T>>& grid) {
          (std::array<double,4> (Gr::*)(const Fractional&) const)
          &Gr::tricubic_interpolation_der)
     .def("interpolate_position_array",
-         [](const Gr& self, const nb::ndarray<double, nb::shape<-1,3>, nb::device::cpu>& xyz,
+         [](const Gr& self, const nb::ndarray<const double, nb::shape<-1,3>, nb::device::cpu>& xyz,
             int order, const Transform* to_frac) {
         auto xyz_view = xyz.view();
         size_t len = xyz_view.shape(0);
@@ -323,6 +323,7 @@ void add_grid(nb::module_& m) {
     .def_rw("constant_r", &SolventMasker::constant_r)
     .def_rw("ignore_hydrogen", &SolventMasker::ignore_hydrogen)
     .def_rw("ignore_zero_occupancy_atoms", &SolventMasker::ignore_zero_occupancy_atoms)
+    .def_rw("use_atom_occupancy", &SolventMasker::use_atom_occupancy)
     .def("set_radii", &SolventMasker::set_radii,
          nb::arg("choice"), nb::arg("constant_r")=0.)
     .def("put_mask_on_int8_grid", &SolventMasker::put_mask_on_grid<int8_t>)
