@@ -219,6 +219,11 @@ void add_software(Metadata& meta, SoftwareItem::Classification type, const std::
     size_t sep = item.name.find(' ');
     if (sep != std::string::npos) {
       size_t ver_start = item.name.find_first_not_of(" (", sep + 1);
+      if (ver_start == std::string::npos) {
+        item.name.resize(sep);
+        item.classification = type;
+        continue;
+      }
       item.version = item.name.substr(ver_start);
       item.name.resize(sep);
       if (!item.version.empty() && item.version.back() == ')') {
@@ -857,11 +862,11 @@ void read_metadata_from_remarks(Structure& st) {
 
 } // anonymous namespace
 
-Structure read_pdb_from_stream(AnyStream& line_reader, const std::string& source,
-                               PdbReadOptions options) {
+void populate_structure_from_pdb_stream(AnyStream& line_reader, const std::string& source,
+                                        Structure& st,
+                                        PdbReadOptions options) {
   if (options.max_line_length <= 0 || options.max_line_length > 120)
     options.max_line_length = 120;
-  Structure st;
   st.input_format = CoorFormat::Pdb;
   st.name = path_basename(source, {".gz", ".pdb"});
   Transform matrix;
@@ -878,6 +883,12 @@ Structure read_pdb_from_stream(AnyStream& line_reader, const std::string& source
   };
   while (size_t len = line_reader.copy_line(line, options.max_line_length+1)) {
     ++line_num;
+    if (options.check_non_ascii && st.non_ascii_line == 0)
+      for (size_t i = 0; i < len; ++i)
+        if (static_cast<unsigned char>(line[i]) >= 0x80) {
+          st.non_ascii_line = line_num;
+          break;
+        }
     if (is_record_type4(line, "ATOM") || is_record_type4(line, "HETATM")) {
       if (len < 55)
         wrong("The line is too short to be correct:\n" + std::string(line));
@@ -983,7 +994,7 @@ Structure read_pdb_from_stream(AnyStream& line_reader, const std::string& source
       std::string chain_name = read_string(line+10, 2);
       Entity& ent = impl::find_or_add(st.entities, chain_name);
       ent.entity_type = EntityType::Polymer;
-      for (int i = 19; i < 68; i += 4) {
+      for (int i = 19; i < 68 && i < (int)len; i += 4) {
         std::string res_name = read_string(line+i, 3);
         if (!res_name.empty())
           ent.full_sequence.emplace_back(res_name);
@@ -1236,7 +1247,6 @@ Structure read_pdb_from_stream(AnyStream& line_reader, const std::string& source
     read_metadata_from_remarks(st);
 
   restore_full_ccd_codes(st);
-  return st;
 }
 
 std::vector<Op> read_remark_290(const std::vector<std::string>& raw_remarks) {
